// PClient.cpp: implementation of the PClient class.
/*! \file PClient.cpp
*/
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "PClient.h"
#include "errorCodes.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!
 * Constructor 
 *
 * @param name : A TString from which the pipe name is derived
 * @param isOut : Boolean, true if this is an outgoing pipe
 */
PClient::PClient(const TString &name, const bool isOut)
: Pipe( name, isOut )
{
	hPipe = INVALID_HANDLE_VALUE;
}

/*!
 * Object Destructor
 */
PClient::~PClient()
{
	disconnect();
}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * Disconnects the pipe and resets the servers event (signaling it 
 * to disconnect).
 */
void PClient::disconnect()
{	
	if(hPipe != INVALID_HANDLE_VALUE){
		CloseHandle(hPipe);
		hPipe = INVALID_HANDLE_VALUE;
	}
	ResetEvent(serverStayConnected);
}

/*!
 * Attempts to connect to server, fails if server isn't serving. 
 */
DWORD PClient::connect()
{
	if( isConnected() )
		return PC_ALREADY_CONNECTED;
	
	if( WaitForSingleObject(clientStayConnected, 0) == WAIT_OBJECT_0 ){
		return PC_NOT_RESET;	// Server yet to process disconnect event
	}

	hPipe = CreateFile( 
		pipeName,			// pipe name 
		isOutbound ? GENERIC_WRITE : GENERIC_READ,	// read and write access 
		0,					// no sharing 
		NULL,				// no security attributes
		OPEN_EXISTING,		// opens existing pipe 
		0,					// default attributes 
		NULL);				// no template file 
	
	if (hPipe == INVALID_HANDLE_VALUE){ 
		if (GetLastError() == ERROR_PIPE_BUSY) 
			return PC_BUSY;
		else
			return PC_FAILED;
	} else
		return PC_CONNECTED;
}

/*!
 * \brief Returns true if the server is connected to a client.
 * 
 * The makes sure the client is not wanting to disconnect, disconnecting as 
 * necessary.  This should be called each time before a pipe operation is
 * conducted.
 */
bool PClient::isConnected()
{
	if( WaitForSingleObject(serverStayConnected, 0) == WAIT_OBJECT_0 ){
		this->disconnect();
	}

	return 	hPipe != INVALID_HANDLE_VALUE;
}
