// PServer.cpp: implementation of the PServer class.
/*! \file PServer.cpp 
*/
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PServer.h"
#include "errorCodes.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!
 * Constructor 
 *
 * @param name : A TString from which the pipe name is derived
 * @param isOut : Boolean, true if this is an outgoing pipe
 * @param bufferSize : Size of the outgoing buffer.  MS documentation seems to be
 * inacurate and the actual buffer size allocated is a factor of this value
 */
PServer::PServer(const TString &name, const bool isOut, const DWORD bufferSize)
: Pipe( name, isOut )
{
	hConnectThread = NULL;
	connected = false;

	if( isOutbound ){
		in = 0;
		out = bufferSize;
	} else {
		in = bufferSize;
		out = 0;
	}
}

/*!
 * Object Destructor
 */
PServer::~PServer()
{
	disconnect();
	if(hConnectThread != NULL)
		CloseHandle(hConnectThread);
}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * \brief Disconnects the pipe from the client.  
 *
 * Disconnects the server end of a pipe and resets the clients stay connected 
 * event. If the pipe is still waiting for a client the thread is terminated.
 */
void PServer::disconnect()
{
	if( getAttemptConnectResult() == PS_WAITING ) {
		/*! \bug ~4Kb memory leak results from terminating a connecting thread */
		TerminateThread(hConnectThread, PS_CANCELLED);  
		connected = false;
	}
	
	connected = false;
	DisconnectNamedPipe(hPipe);
	CloseHandle(hPipe);
	
	ResetEvent(clientStayConnected);
}

/*!
 * Serves a Pipe and starts a thread to wait for a client
 *
 * @return bool : true iff pipe is open for clients
 */
bool PServer::serve()
{
	if( PServer::isConnected() ){
		return true;
	}

	if( WaitForSingleObject(serverStayConnected, 0) == WAIT_OBJECT_0 ){
		return false; // Client yet to process disconnect request
	}

	hPipe = CreateNamedPipe(
		pipeName,			// pipe name
		isOutbound ? PIPE_ACCESS_OUTBOUND : PIPE_ACCESS_INBOUND,	// pipe open mode
		0,					// Byte mode, blocking operations
		1,					// Maximum pipe Instances
		out,				// output buffer size
		in,					// input buffer size
		NMPWAIT_WAIT_FOREVER,	// default time-out interval
		NULL);				// Default Security, non inheritable

	if( GetLastError() == ERROR_ALREADY_EXISTS ) {
		CloseHandle(hPipe);  // Close duplicated handle
		hPipe = INVALID_HANDLE_VALUE;
	}
	if ( hPipe == INVALID_HANDLE_VALUE ) {
		return false;
	}
	
	if(hConnectThread != NULL)
		CloseHandle(hConnectThread);

	hConnectThread = chBEGINTHREADEX(
		NULL,			// Default Security, non inheritable
		0,				// use default stack size  
		&waitForClient,	// thread function
		this,			// thread argument
		0,				// use default creation flags 
		NULL);			// don't need thread identifier

	if(hConnectThread == NULL){
		CloseHandle(hPipe);
		return false;
	}

	return true;
}

/*!
 * Thread function which is blocked until client connects
 */
DWORD WINAPI PServer::waitForClient(LPVOID param)
{
	PServer* server = (PServer*) param;

	// Wait for the client to connect; if it succeeds, 
	// the function returns a nonzero value. If the function returns 
	// zero, GetLastError returns ERROR_PIPE_CONNECTED.
 	server->connected = ConnectNamedPipe(server->hPipe, NULL) ? 
		true : (GetLastError() == ERROR_PIPE_CONNECTED);

	if( server->connected ){
		return PS_CONNECTED;
	} else {
		return PS_FAILED;
	}

}

/*!
 * Produces a handle to an Event Obect that becomes sigalled when a client connects
 */
HANDLE PServer::getConnectedEventHandle()
{
	HANDLE rtrn;

	// The thread handle can be used as an event object.
	// becomes signalled when the thread stops running

	DuplicateHandle(
		GetCurrentProcess(),
		hConnectThread,		// handle to duplicate
		GetCurrentProcess(),
		&rtrn,				// duplicate handle
		SYNCHRONIZE,		// handle can only be used to wait on
		false,				// non inheritable
		0);					// no extra options

	return rtrn;
}

/*!
 * Returns the result of a connection attept
 */
DWORD PServer::getAttemptConnectResult() const
{
	DWORD exitCode;
	GetExitCodeThread(hConnectThread, &exitCode);

	return exitCode;
}

/*!
 * \brief Returns true if the server is connected to a client.
 * 
 * The makes sure the client is not wanting to disconnect, disconnecting as 
 * necessary.  This should be called each time before a pipe operation is
 * conducted.
 */
bool PServer::isConnected()
{
	if( WaitForSingleObject(clientStayConnected, 0) == WAIT_OBJECT_0 ){
		this->disconnect();
	}

	return connected;
}
