// WaitCallback.cpp: implementation of the WaitCallback class.
/*! \file WaitCallback.cpp
*/
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "WaitCallback.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!
 * Default Constructor
 */
WaitCallback::WaitCallback()
{
	waitEvents[0] = CreateEvent(NULL, true, false, NULL);
	waitEvents[1] = INVALID_HANDLE_VALUE;	// Important, see hasWaitRegistered
}

/*!
 * Object Destructor
 */
WaitCallback::~WaitCallback()
{
	unregisterWait();
	CloseHandle(waitEvents[0]);
	CloseHandle(waitEvents[1]);
}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * Returns true is there is a registered wait (completed or not)
 */
bool WaitCallback::hasWaitRegistered() const
{
	return waitEvents[1] != INVALID_HANDLE_VALUE;
}

/*!
 * Registers a wait callback.
 *
 * @param hEvent : A waitable kernel object (note that this class 
 * takes possesion so pass a duplicated handle if it is required after 
 * this call.
 * @param cB : A pointer to a WAITCALLBACK function
 * @param cBP : A void pointer passed as the callbacks first parameter
 * @param t0 : The wait timeout (ms) can be INFINITE
 * @param once : True if should Wait and callback just once
 *
 * @return bool : true iff wait successfully registered
 * \note the wait object is not possesed if this function fails (ie 
 * returns false)
 */
bool WaitCallback::registerWaitForSingleObject(
	HANDLE hEvent,		// handle to wait object (takes possesion)
	const WAITCALLBACK cB,	// callback function		
	const PVOID cBP,	// callback context 
	const DWORD tO,		// time-out interval
	const bool once)	// execute just once?
{
	if( hasWaitRegistered() )
		return false;

	hWaitThread = chBEGINTHREADEX(
		NULL,			// Default Security, non inheritable
		0,				// use default stack size  
		&waitThread,	// thread function
		this,			// thread argument
		CREATE_SUSPENDED,	// Create suspended
		NULL);			// don't need thread identifier

	if(hWaitThread == NULL)
		return false;

	// Take possesion of the event handle
	waitEvents[1] = hEvent;
	cBack = cB;
	cBackParam = cBP;
	timeOut = tO;
	justOnce = once;

	ResumeThread(hWaitThread);

	return true;
}

/*!
 * Unregister the wait blocks for upto 6 seconds and then forces termination
 */
void WaitCallback::unregisterWait()
{
	if( hasWaitRegistered() ){
		SetEvent(waitEvents[0]);	// Signal waitThread to exit

		if( WaitForSingleObject(hWaitThread, 6000) == WAIT_TIMEOUT ){
			TerminateThread(hWaitThread, 0);
			chMB("Warning couldn't unregister a wait gracefully.\nCheck for Deadlock");
		}

		CloseHandle(waitEvents[1]);
		waitEvents[1] = INVALID_HANDLE_VALUE;
		CloseHandle(hWaitThread);
		ResetEvent(waitEvents[0]);
	}
}

/*!
 * The thread funcion which performs the wait and calls the callback function
 */
DWORD WINAPI WaitCallback::waitThread(LPVOID param)
{
	WaitCallback* thisWait = (WaitCallback*) param;

	do{
		switch( WaitForMultipleObjects(2, thisWait->waitEvents, false, thisWait->timeOut) ) {
		case WAIT_OBJECT_0:	// wait is being unregistered
			return true;
			break;
		case WAIT_FAILED:
			assert(WAIT_FAILED != WAIT_FAILED);
			return false;
			break;
		case WAIT_TIMEOUT:
			(*thisWait->cBack)(thisWait->cBackParam, true);
			break;
		default:
			(*thisWait->cBack)(thisWait->cBackParam, false);
			break;
		}

	} while ( !thisWait->justOnce );
	
	return true;
}


