// BitmapScreen.cpp: implementation of the BitmapScreen class.
/*! \file BitmapScreen.cpp
*/
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "BitmapScreen.h"

//////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////

/*!
 * Constructor - Takes the same parameters as VScreen
 */
BitmapScreen::BitmapScreen(const int tlX, const int tlY, const int w, const int h)
: VScreen(tlX, tlY, w, h)
{
	currentBuffer = 0;
	buffer[0].createCompatible(w, h, 8);
	buffer[1].createCompatible(w, h, 8);

	writeMutex = CreateMutex(NULL, false, NULL);
}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * Requests a read lock and then sets the dibs bits to the specified DC.
 * Releases the read lock before returning.
 * @param hDC : Handle to a device context.
 */
void BitmapScreen::paint(HDC hDC)
{
	currentBufferLock.WaitToRead();

	RECT regionR = VScreen::getRgnRect();

	SetDIBitsToDevice(
		hDC,
		regionR.left, regionR.top,
		buffer[currentBuffer].getPBITMAPINFOHEADER()->biWidth,
		abs(buffer[currentBuffer].getPBITMAPINFOHEADER()->biHeight),
		0, 0,
		0,
		abs(buffer[currentBuffer].getPBITMAPINFOHEADER()->biHeight),
		buffer[currentBuffer].getPBytes(),
		buffer[currentBuffer].getPBITMAPINFO(),
		DIB_RGB_COLORS);

	currentBufferLock.Done();
}

/*! 
 * Obtains a read lock and returns a pointer to a const DIBitmap.
 * The pointer should be considered invalid after calling DoneReading.
 */
const DIBitmap* BitmapScreen::getReadBuffer()
{
	currentBufferLock.WaitToRead();
	return &buffer[currentBuffer];
}

/*!
 * Releases the read lock. This most be called for each read lock obtained
 * or a write lock will never suceed and block any other lock attempts.
 */
void BitmapScreen::DoneReading()
{
	currentBufferLock.Done();
}

/*!
 * Returns a pointer to a writable DIBitmap.  This function is serialised
 * using the writeMutex. The pointer should be considered invalid after 
 * calling DoneWritting.
 */
DIBitmap* BitmapScreen::getWriteBuffer()
{
	WaitForSingleObject(writeMutex, INFINITE);
	return &buffer[0 == currentBuffer ? 1 : 0];
}

/*!
 * After checking the validity of the updated buffer a write lock is obtained
 * the buffers are switched and the lock released.  The Mutex serialising 
 * writes is also released.
 *
 * @return bool : true iff the buffer written to is a bitmap compatible with 
 * this screen
 */
bool BitmapScreen::DoneWritting()
{
	int otherBuffer = 0 == currentBuffer ? 1 : 0;
	
	if(	abs(buffer[otherBuffer].getPBITMAPINFOHEADER()->biHeight) != abs(VScreen::getHeight())
			|| buffer[otherBuffer].getPBITMAPINFOHEADER()->biWidth != VScreen::getWidth() ){
		assert( ReleaseMutex(writeMutex) );
		return false;
	}

	currentBufferLock.WaitToWrite();

	currentBuffer = otherBuffer;

	currentBufferLock.Done();

	assert( ReleaseMutex(writeMutex) );
	return true;
}


