// MousePipe.cpp: implementation of the MousePipe class.
/*! \file MousePipe.cpp
*/
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MousePipe.h"

/*! 
 * For some reason I couldn't get buffer sizes to play nice 
 * (maybe the data written to the pipe consists of more than
 * the actual message).
 */
#define BUFFER_LIES_INT_FACTOR 7	

/*!
 * The maximum number of messages that can be queued.
 */
#define QUEUE_LENGTH 4 

/*!
 * The size of the buffer required to store the queue on the pipe.
 */
#define BUFFER_SIZE sizeof(mouseMessage) * QUEUE_LENGTH * BUFFER_LIES_INT_FACTOR

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!
 * @param name : A TString used to generate unique names for the
 * various kernel objects
 */
MousePipe::MousePipe(const TString &name)
:PServer(TString( TEXT("Mouse") ) + name, true, BUFFER_SIZE)
{
	semName = TString( TEXT("MOUSESEM") ) + name;
	numberInQueue = NULL;
}

/*!
 * Object Destructor
 */
MousePipe::~MousePipe()
{
	if(numberInQueue != NULL)
		CloseHandle(numberInQueue);
}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * Queues a mouseMessage to the pipe
 * @param x : The x co-ordinate
 * @param y : The y co-ordinate
 * @param buttonState : 0 = button down, 1 = button up
 * @return bool : true iff message was queued to the pipe
 */
bool MousePipe::queueMessage(const int x, const int y, const int buttonState)
{
    if( !isConnected() ){
		return false;
    }

    mouseMessage msg;
    msg.x = x;
    msg.y = y;
	msg.button = buttonState;

    DWORD bytesWritten;
    if( ReleaseSemaphore(numberInQueue, 1, NULL) ){	
		// WriteFile will return very quickly since there is space in the buffer
        return WriteFile(hPipe, &msg, sizeof(mouseMessage), &bytesWritten, NULL)
			? bytesWritten == sizeof(mouseMessage) : false;
    } else { // no space in buffer
        return false;
    }
}

/*!
 * Calls PServer::serve() and if successful creates a semaphore.  See
 * documentation for PServer.
 */
bool MousePipe::serve()
{
	if( !PServer::serve() ) {
		return false;
	} else {
		numberInQueue = CreateSemaphore(
			NULL,           // default security
			0,             // initial count ie 0 messages in pipe
			QUEUE_LENGTH,   // maximum count
			semName);

		if(numberInQueue == NULL){
			disconnect();
			return false;
		}

		return true;
	}	
}

/*!
 * Calls PServer::disconnect() and removes the semaphore.  See
 * documentation for PServer.
 */
void MousePipe::disconnect()
{	
	PServer::disconnect();

	if(numberInQueue != NULL)
		CloseHandle(numberInQueue);
}