// ScreenPipe.cpp: implementation of the ScreenPipe class.
/*! \file ScreenPipe.cpp
*/
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ScreenPipe.h"

//! A factor required to ensure the pipe buffer is large enough
#define BUFFER_LIES_BYTE_FACTOR 23

/*! The Size of a buffer required to write a 240*320 8bpp bitmap to the pipe.
	Fitting the entire image to pipe is not critical.*/
#define BUFFER_SIZE sizeof(BYTE)*240*320+(sizeof(BYTE)*3*256)+sizeof(screenInfo) \
	* BUFFER_LIES_BYTE_FACTOR

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

/*!
 * @param name : A TString used to generate unique names for the
 * various kernel objects
 */
ScreenPipe::ScreenPipe(const TString &name)
:PServer(TString( TEXT("Screen") ) + name, 
	false, 
	BUFFER_SIZE)
{}

//////////////////////////////////////////////////////////////////////
// Member Functions
//////////////////////////////////////////////////////////////////////

/*!
 * A blocking function which reads data from the screen pipe and 
 * sets the given DIBitmap accordingly.
 * @param destination : The DIBitmap which is set
 */
DWORD ScreenPipe::readToDIB(DIBitmap &destination)
{
	DWORD numBytesRead;
	bool isOk;

	screenInfo header;
	isOk = ReadFile( 
				hPipe,					// Read from
				&header,				// data buffer
				sizeof(screenInfo),		// number of bytes to read
				&numBytesRead,			// number of bytes read
				NULL);					// not overlapped
	
	if( !isOk || (numBytesRead != sizeof(screenInfo)) )
		return readError();

	if( header.hdrSize != sizeof(screenInfo) ){
		return SP_INVALID_DATA;
	}

	//! if destination isn't a compatible bitmap reset it
	if( !(destination.getPBITMAPINFOHEADER()->biBitCount == header.bitCount
			&& destination.getPBITMAPINFOHEADER()->biHeight == header.height
			&& destination.getPBITMAPINFOHEADER()->biWidth == header.width) ){

		if( !destination.createCompatible(header.width, header.height, header.bitCount))
			return SP_ERROR_CREATING_BITMAP;
	}

	LPRGBQUAD palette = destination.getPPalette();
	RGBTRIPLE entry;
	for(int i=0; i < header.clrUsed; i++, palette++){

		isOk = ReadFile(hPipe, &entry, sizeof(RGBTRIPLE), &numBytesRead, NULL);
		if( !isOk || (numBytesRead != sizeof(RGBTRIPLE)) )
			return readError();

		palette->rgbBlue = entry.rgbtBlue;
		palette->rgbGreen = entry.rgbtGreen;
		palette->rgbRed = entry.rgbtRed;
		palette->rgbReserved = 0;
	}

	// See DIBitmap documentation for explanation of this formula
	const DWORD size = (((( header.width * header.bitCount ) + 31) & ~31) >> 3) * header.height;

	isOk = ReadFile(hPipe, destination.getPBytes(), size, &numBytesRead, NULL);
	if( !isOk || (numBytesRead != size ) )
		return readError();

	return SP_SUCCESS;
}

/*!
 * Error function for readToDIB()
 */
DWORD ScreenPipe::readError()
{
	switch(GetLastError()) {
		case ERROR_SECTOR_NOT_FOUND:
			disconnect();
			return SP_NOT_CONNECTED;
		case ERROR_SUCCESS:		// EOF?
		case ERROR_READ_FAULT:
		default:
			return SP_UNKNOWN_ERROR;
	}
}
